__all__ = ["ApiGroup"]

from typing import Callable, Optional, TypeVar

from arango.connection import Connection
from arango.executor import ApiExecutor
from arango.request import Request
from arango.response import Response
from arango.result import Result

T = TypeVar("T")


class ApiGroup:
    """Base class for API groups.

    :param connection: HTTP connection.
    :param executor: API executor.
    """

    def __init__(self, connection: Connection, executor: ApiExecutor) -> None:
        self._conn = connection
        self._executor = executor

    @property
    def conn(self) -> Connection:
        """Return the HTTP connection.

        :return: HTTP connection.
        :rtype: arango.connection.BasicConnection | arango.connection.JwtConnection |
            arango.connection.JwtSuperuserConnection
        """
        return self._conn

    @property
    def db_name(self) -> str:
        """Return the name of the current database.

        :return: Database name.
        :rtype: str
        """
        return self._conn.db_name

    @property
    def username(self) -> Optional[str]:
        """Return the username.

        :returns: Username.
        :rtype: str
        """
        return self._conn.username

    @property
    def context(self) -> str:
        """Return the API execution context.

        :return: API execution context. Possible values are "default", "async",
            "batch" and "transaction".
        :rtype: str
        """
        return self._executor.context

    def _execute(
        self, request: Request, response_handler: Callable[[Response], T]
    ) -> Result[T]:
        """Execute an API.

        :param request: HTTP request.
        :type request: arango.request.Request
        :param response_handler: HTTP response handler.
        :type response_handler: callable
        :return: API execution result.
        """
        return self._executor.execute(request, response_handler)



project = "python-arango"
copyright = "2016-2022, Joohwan Oh"
author = "Joohwan Oh"
extensions = [
    "sphinx_rtd_theme",
    "sphinx.ext.autodoc",
    "sphinx.ext.doctest",
    "sphinx.ext.viewcode",
]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]
html_static_path = ["static"]
html_theme = "sphinx_rtd_theme"
master_doc = "index"

autodoc_member_order = "bysource"

doctest_global_setup = """
from arango import ArangoClient
# Initialize the ArangoDB client.
client = ArangoClient()
# Connect to "_system" database as root user.
sys_db = client.db('_system', username='root', password='passwd')
# Create "test" database if it does not exist.
if not sys_db.has_database('test'):
    sys_db.create_database('test')
# Ensure that user "johndoe@gmail.com" does not exist.
if sys_db.has_user('johndoe@gmail.com'):
    sys_db.delete_user('johndoe@gmail.com')
# Connect to "test" database as root user.
db = client.db('test', username='root', password='passwd')
# Create "students" collection if it does not exist.
if db.has_collection('students'):
    db.collection('students').truncate()
else:
    db.create_collection('students')
# Ensure that "cities" collection does not exist.
if db.has_collection('cities'):
    db.delete_collection('cities')
# Create "school" graph if it does not exist.
if db.has_graph("school"):
    school = db.graph('school')
else:
    school = db.create_graph('school')
# Create "teachers" vertex collection if it does not exist.
if school.has_vertex_collection('teachers'):
    school.vertex_collection('teachers').truncate()
else:
    school.create_vertex_collection('teachers')
# Create "lectures" vertex collection if it does not exist.
if school.has_vertex_collection('lectures'):
    school.vertex_collection('lectures').truncate()
else:
    school.create_vertex_collection('lectures')
# Create "teach" edge definition if it does not exist.
if school.has_edge_definition('teach'):
    school.edge_collection('teach').truncate()
else:
    school.create_edge_definition(
        edge_collection='teach',
        from_vertex_collections=['teachers'],
        to_vertex_collections=['lectures']
    )
"""



__all__ = [
    "HostResolver",
    "SingleHostResolver",
    "RandomHostResolver",
    "RoundRobinHostResolver",
]

import random
from abc import ABC, abstractmethod
from typing import Optional, Set


class HostResolver(ABC):  # pragma: no cover
    """Abstract base class for host resolvers."""

    def __init__(self, host_count: int = 1, max_tries: Optional[int] = None) -> None:
        max_tries = max_tries or host_count * 3
        if max_tries < host_count:
            raise ValueError("max_tries cannot be less than host_count")

        self._host_count = host_count
        self._max_tries = max_tries

    @abstractmethod
    def get_host_index(self, indexes_to_filter: Optional[Set[int]] = None) -> int:
        raise NotImplementedError

    @property
    def host_count(self) -> int:
        return self._host_count

    @property
    def max_tries(self) -> int:
        return self._max_tries


class SingleHostResolver(HostResolver):
    """Single host resolver."""

    def get_host_index(self, indexes_to_filter: Optional[Set[int]] = None) -> int:
        return 0


class RandomHostResolver(HostResolver):
    """Random host resolver."""

    def __init__(self, host_count: int, max_tries: Optional[int] = None) -> None:
        super().__init__(host_count, max_tries)

    def get_host_index(self, indexes_to_filter: Optional[Set[int]] = None) -> int:
        host_index = None
        indexes_to_filter = indexes_to_filter or set()
        while host_index is None or host_index in indexes_to_filter:
            host_index = random.randint(0, self.host_count - 1)

        return host_index


class RoundRobinHostResolver(HostResolver):
    """Round-robin host resolver."""

    def __init__(self, host_count: int, max_tries: Optional[int] = None) -> None:
        super().__init__(host_count, max_tries)
        self._index = -1

    def get_host_index(self, indexes_to_filter: Optional[Set[int]] = None) -> int:
        self._index = (self._index + 1) % self.host_count
        return self._index



__all__ = [
    "ApiExecutor",
    "DefaultApiExecutor",
    "AsyncApiExecutor",
    "BatchApiExecutor",
    "TransactionApiExecutor",
]

from collections import OrderedDict
from typing import Any, Callable, Optional, Sequence, Tuple, TypeVar, Union
from urllib.parse import urlencode
from uuid import uuid4

from arango.connection import Connection
from arango.exceptions import (
    AsyncExecuteError,
    BatchExecuteError,
    BatchStateError,
    TransactionAbortError,
    TransactionCommitError,
    TransactionInitError,
    TransactionStatusError,
)
from arango.job import AsyncJob, BatchJob
from arango.request import Request
from arango.response import Response
from arango.typings import Fields, Json
from arango.utils import suppress_warning

ApiExecutor = Union[
    "DefaultApiExecutor",
    "AsyncApiExecutor",
    "BatchApiExecutor",
    "TransactionApiExecutor",
]

T = TypeVar("T")


class DefaultApiExecutor:
    """Default API executor.

    :param connection: HTTP connection.
    :type connection: arango.connection.BasicConnection |
        arango.connection.JwtConnection | arango.connection.JwtSuperuserConnection
    """

    def __init__(self, connection: Connection) -> None:
        self._conn = connection

    @property
    def context(self) -> str:
        return "default"

    def execute(self, request: Request, response_handler: Callable[[Response], T]) -> T:
        """Execute an API request and return the result.

        :param request: HTTP request.
        :type request: arango.request.Request
        :param response_handler: HTTP response handler.
        :type response_handler: callable
        :return: API execution result.
        """
        resp = self._conn.send_request(request)
        return response_handler(resp)


class AsyncApiExecutor:
    """Async API Executor.

    :param connection: HTTP connection.
    :type connection: arango.connection.BasicConnection |
        arango.connection.JwtConnection | arango.connection.JwtSuperuserConnection
    :param return_result: If set to True, API executions return instances of
        :class:`arango.job.AsyncJob` and results can be retrieved from server
        once available. If set to False, API executions return None and no
        results are stored on server.
    :type return_result: bool
    """

    def __init__(self, connection: Connection, return_result: bool) -> None:
        self._conn = connection
        self._return_result = return_result

    @property
    def context(self) -> str:
        return "async"

    def execute(
        self, request: Request, response_handler: Callable[[Response], T]
    ) -> Optional[AsyncJob[T]]:
        """Execute an API request asynchronously.

        :param request: HTTP request.
        :type request: arango.request.Request
        :param response_handler: HTTP response handler.
        :type response_handler: callable
        :return: Async job or None if **return_result** parameter was set to
            False during initialization.
        :rtype: arango.job.AsyncJob | None
        """
        if self._return_result:
            request.headers["x-arango-async"] = "store"
        else:
            request.headers["x-arango-async"] = "true"

        resp = self._conn.send_request(request)
        if not resp.is_success:
            raise AsyncExecuteError(resp, request)
        if not self._return_result:
            return None

        job_id = resp.headers["x-arango-async-id"]
        return AsyncJob(self._conn, job_id, response_handler)


class BatchApiExecutor:
    """Batch API executor.

    :param connection: HTTP connection.
    :param return_result: If set to True, API executions return instances of
        :class:`arango.job.BatchJob` that are populated with results on commit.
        If set to False, API executions return None and no results are tracked
        client-side.
    :type return_result: bool
    """

    def __init__(self, connection: Connection, return_result: bool) -> None:
        self._conn = connection
        self._return_result: bool = return_result
        self._queue: OrderedDict[str, Tuple[Request, BatchJob[Any]]] = OrderedDict()
        self._committed: bool = False

    @property
    def context(self) -> str:
        return "batch"

    def _stringify_request(self, request: Request) -> str:
        path = request.endpoint

        if request.params is not None:
            path += f"?{urlencode(request.params)}"
        buffer = [f"{request.method} {path} HTTP/1.1"]

        if request.headers is not None:
            for key, value in sorted(request.headers.items()):
                buffer.append(f"{key}: {value}")

        if request.data is not None:
            serialized = self._conn.serialize(request.data)
            buffer.append("\r\n" + serialized)

        return "\r\n".join(buffer)

    @property
    def jobs(self) -> Optional[Sequence[BatchJob[Any]]]:
        """Return the queued batch jobs.

        :return: Batch jobs or None if **return_result** parameter was set to
            False during initialization.
        :rtype: [arango.job.BatchJob] | None
        """
        if not self._return_result:
            return None
        return [job for _, job in self._queue.values()]

    def execute(
        self, request: Request, response_handler: Callable[[Response], T]
    ) -> Optional[BatchJob[T]]:
        """Place the request in the batch queue.

        :param request: HTTP request.
        :type request: arango.request.Request
        :param response_handler: HTTP response handler.
        :type response_handler: callable
        :return: Batch job or None if **return_result** parameter was set to
            False during initialization.
        :rtype: arango.job.BatchJob | None
        :raise arango.exceptions.BatchStateError: If batch was already
            committed.
        """
        if self._committed:
            raise BatchStateError("batch already committed")

        job = BatchJob(response_handler)
        self._queue[job.id] = (request, job)
        return job if self._return_result else None

    def commit(self) -> Optional[Sequence[BatchJob[Any]]]:
        """Execute the queued requests in a single batch API request.

        If **return_result** parameter was set to True during initialization,
        :class:`arango.job.BatchJob` instances are populated with results.

        :return: Batch jobs or None if **return_result** parameter was set to
            False during initialization.
        :rtype: [arango.job.BatchJob] | None
        :raise arango.exceptions.BatchStateError: If batch state is invalid
            (e.g. batch was already committed or size of response from server
            did not match the expected).
        :raise arango.exceptions.BatchExecuteError: If commit fails.
        """
        if self._committed:
            raise BatchStateError("batch already committed")
        else:
            self._committed = True

        if len(self._queue) == 0:
            return self.jobs

        # Boundary used for multipart request
        boundary = uuid4().hex

        # Build the batch request payload
        buffer = []
        for req, job in self._queue.values():
            buffer.append(f"--{boundary}")
            buffer.append("Content-Type: application/x-arango-batchpart")
            buffer.append(f"Content-Id: {job.id}")
            buffer.append("\r\n" + self._stringify_request(req))
        buffer.append(f"--{boundary}--")

        request = Request(
            method="post",
            endpoint="/_api/batch",
            headers={"Content-Type": f"multipart/form-data; boundary={boundary}"},
            data="\r\n".join(buffer),
        )
        with suppress_warning("requests.packages.urllib3.connectionpool"):
            resp = self._conn.send_request(request)

        if not resp.is_success:
            raise BatchExecuteError(resp, request)

        if not self._return_result:
            return None

        url_prefix = resp.url.strip("/_api/batch")
        raw_responses = resp.raw_body.split(f"--{boundary}")[1:-1]

        if len(self._queue) != len(raw_responses):
            raise BatchStateError(
                "expecting {} parts in batch response but got {}".format(
                    len(self._queue), len(raw_responses)
                )
            )
        for raw_resp in raw_responses:
            # Parse and breakdown the batch response body
            resp_parts = raw_resp.strip().split("\r\n")
            raw_content_id = resp_parts[1]
            raw_body = resp_parts[-1]
            raw_status = resp_parts[3]
            job_id = raw_content_id.split(" ")[1]
            _, status_code, status_text = raw_status.split(" ", 2)

            # Update the corresponding batch job
            queued_req, queued_job = self._queue[job_id]

            queued_job._status = "done"
            resp = Response(
                method=queued_req.method,
                url=url_prefix + queued_req.endpoint,
                headers={},
                status_code=int(status_code),
                status_text=status_text,
                raw_body=raw_body,
            )
            queued_job._response = self._conn.prep_response(resp)

        return self.jobs


class TransactionApiExecutor:
    """Executes transaction API requests.

    :param connection: HTTP connection.
    :param read: Name(s) of collections read during transaction. Read-only
        collections are added lazily but should be declared if possible to
        avoid deadlocks.
    :type read: str | [str]
    :param write: Name(s) of collections written to during transaction with
        shared access.
    :type write: str | [str]
    :param exclusive: Name(s) of collections written to during transaction
        with exclusive access.
    :type exclusive: str | [str]
    :param sync: Block until operation is synchronized to disk.
    :type sync: bool | None
    :param allow_implicit: Allow reading from undeclared collections.
    :type allow_implicit: bool
    :param lock_timeout: Timeout for waiting on collection locks. If not given,
        a default value is used. Setting it to 0 disables the timeout.
    :type lock_timeout: int
    :param max_size: Max transaction size in bytes.
    :type max_size: int
    """

    def __init__(
        self,
        connection: Connection,
        read: Optional[Fields] = None,
        write: Optional[Fields] = None,
        exclusive: Optional[Fields] = None,
        sync: Optional[bool] = None,
        allow_implicit: Optional[bool] = None,
        lock_timeout: Optional[int] = None,
        max_size: Optional[int] = None,
    ) -> None:
        self._conn = connection

        collections: Json = {}
        if read is not None:
            collections["read"] = read
        if write is not None:
            collections["write"] = write
        if exclusive is not None:
            collections["exclusive"] = exclusive

        data: Json = {"collections": collections}
        if sync is not None:
            data["waitForSync"] = sync
        if allow_implicit is not None:
            data["allowImplicit"] = allow_implicit
        if lock_timeout is not None:
            data["lockTimeout"] = lock_timeout
        if max_size is not None:
            data["maxTransactionSize"] = max_size

        request = Request(method="post", endpoint="/_api/transaction/begin", data=data)
        resp = self._conn.send_request(request)

        if not resp.is_success:
            raise TransactionInitError(resp, request)

        result: Json = resp.body["result"]
        self._id: str = result["id"]

    @property
    def context(self) -> str:
        return "transaction"

    @property
    def id(self) -> str:
        """Return the transaction ID.

        :return: Transaction ID.
        :rtype: str
        """
        return self._id

    def execute(self, request: Request, response_handler: Callable[[Response], T]) -> T:
        """Execute API request in a transaction and return the result.

        :param request: HTTP request.
        :type request: arango.request.Request
        :param response_handler: HTTP response handler.
        :type response_handler: callable
        :return: API execution result.
        """
        request.headers["x-arango-trx-id"] = self._id
        resp = self._conn.send_request(request)
        return response_handler(resp)

    def status(self) -> str:
        """Return the transaction status.

        :return: Transaction status.
        :rtype: str
        :raise arango.exceptions.TransactionStatusError: If retrieval fails.
        """
        request = Request(
            method="get",
            endpoint=f"/_api/transaction/{self._id}",
        )
        resp = self._conn.send_request(request)

        if resp.is_success:
            return str(resp.body["result"]["status"])
        raise TransactionStatusError(resp, request)

    def commit(self) -> bool:
        """Commit the transaction.

        :return: True if commit was successful.
        :rtype: bool
        :raise arango.exceptions.TransactionCommitError: If commit fails.
        """
        request = Request(
            method="put",
            endpoint=f"/_api/transaction/{self._id}",
        )
        resp = self._conn.send_request(request)

        if resp.is_success:
            return True
        raise TransactionCommitError(resp, request)

    def abort(self) -> bool:
        """Abort the transaction.

        :return: True if the abort operation was successful.
        :rtype: bool
        :raise arango.exceptions.TransactionAbortError: If abort fails.
        """
        request = Request(
            method="delete",
            endpoint=f"/_api/transaction/{self._id}",
        )
        resp = self._conn.send_request(request)

        if resp.is_success:
            return True
        raise TransactionAbortError(resp, request)



__all__ = ["Cursor"]

from collections import deque
from typing import Any, Deque, Optional, Sequence

from arango.connection import BaseConnection
from arango.exceptions import (
    CursorCloseError,
    CursorCountError,
    CursorEmptyError,
    CursorNextError,
    CursorStateError,
)
from arango.request import Request
from arango.typings import Json


class Cursor:
    """Cursor API wrapper.

    Cursors fetch query results from ArangoDB server in batches. Cursor objects
    are *stateful* as they store the fetched items in-memory. They must not be
    shared across threads without proper locking mechanism.

    :param connection: HTTP connection.
    :param init_data: Cursor initialization data.
    :type init_data: dict
    :param cursor_type: Cursor type ("cursor" or "export").
    :type cursor_type: str
    """

    __slots__ = [
        "_conn",
        "_type",
        "_id",
        "_count",
        "_cached",
        "_stats",
        "_profile",
        "_warnings",
        "_has_more",
        "_batch",
    ]

    def __init__(
        self,
        connection: BaseConnection,
        init_data: Json,
        cursor_type: str = "cursor",
    ) -> None:
        self._conn = connection
        self._type = cursor_type
        self._batch: Deque[Any] = deque()
        self._id = None
        self._count: Optional[int] = None
        self._cached = None
        self._stats = None
        self._profile = None
        self._warnings = None
        self._update(init_data)

    def __iter__(self) -> "Cursor":
        return self

    def __next__(self) -> Any:  # pragma: no cover
        return self.next()

    def __enter__(self) -> "Cursor":
        return self

    def __len__(self) -> int:
        if self._count is None:
            raise CursorCountError("cursor count not enabled")
        return self._count

    def __exit__(self, *_: Any) -> None:
        self.close(ignore_missing=True)

    def __repr__(self) -> str:
        return f"<Cursor {self._id}>" if self._id else "<Cursor>"

    def _update(self, data: Json) -> Json:
        """Update the cursor using data from ArangoDB server.

        :param data: Cursor data from ArangoDB server (e.g. results).
        :type data: dict
        :return: Update cursor data.
        :rtype: dict
        """
        result: Json = {}

        if "id" in data:
            self._id = data["id"]
            result["id"] = data["id"]
        if "count" in data:
            self._count = data["count"]
            result["count"] = data["count"]
        if "cached" in data:
            self._cached = data["cached"]
            result["cached"] = data["cached"]

        self._has_more = bool(data["hasMore"])
        result["has_more"] = data["hasMore"]

        self._batch.extend(data["result"])
        result["batch"] = data["result"]

        if "extra" in data:
            extra = data["extra"]

            if "profile" in extra:
                self._profile = extra["profile"]
                result["profile"] = extra["profile"]

            if "warnings" in extra:
                self._warnings = extra["warnings"]
                result["warnings"] = extra["warnings"]

            if "stats" in extra:
                stats = extra["stats"]
                if "writesExecuted" in stats:
                    stats["modified"] = stats.pop("writesExecuted")
                if "writesIgnored" in stats:
                    stats["ignored"] = stats.pop("writesIgnored")
                if "scannedFull" in stats:
                    stats["scanned_full"] = stats.pop("scannedFull")
                if "scannedIndex" in stats:
                    stats["scanned_index"] = stats.pop("scannedIndex")
                if "executionTime" in stats:
                    stats["execution_time"] = stats.pop("executionTime")
                if "httpRequests" in stats:
                    stats["http_requests"] = stats.pop("httpRequests")
                if "cursorsCreated" in stats:
                    stats["cursorsCreated"] = stats.pop("cursorsCreated")
                if "cursorsRearmed" in stats:
                    stats["cursorsRearmed"] = stats.pop("cursorsRearmed")
                if "cacheHits" in stats:
                    stats["cacheHits"] = stats.pop("cacheHits")
                if "cacheMisses" in stats:
                    stats["cacheMisses"] = stats.pop("cacheMisses")
                self._stats = stats
                result["statistics"] = stats

        return result

    @property
    def id(self) -> Optional[str]:
        """Return the cursor ID.

        :return: Cursor ID.
        :rtype: str
        """
        return self._id

    @property
    def type(self) -> str:
        """Return the cursor type.

        :return: Cursor type ("cursor" or "export").
        :rtype: str
        """
        return self._type

    def batch(self) -> Optional[Deque[Any]]:
        """Return the current batch of results.

        :return: Current batch.
        :rtype: collections.deque
        """
        return self._batch

    def has_more(self) -> Optional[bool]:
        """Return True if more results are available on the server.

        :return: True if more results are available on the server.
        :rtype: bool
        """
        return self._has_more

    def count(self) -> Optional[int]:
        """Return the total number of documents in the entire result set.

        :return: Total number of documents, or None if the count option
            was not enabled during cursor initialization.
        :rtype: int | None
        """
        return self._count

    def cached(self) -> Optional[bool]:
        """Return True if results are cached.

        :return: True if results are cached.
        :rtype: bool
        """
        return self._cached

    def statistics(self) -> Optional[Json]:
        """Return cursor statistics.

        :return: Cursor statistics.
        :rtype: dict
        """
        return self._stats

    def profile(self) -> Optional[Json]:
        """Return cursor performance profile.

        :return: Cursor performance profile.
        :rtype: dict
        """
        return self._profile

    def warnings(self) -> Optional[Sequence[Json]]:
        """Return any warnings from the query execution.

        :return: Warnings, or None if there are none.
        :rtype: [str]
        """
        return self._warnings

    def empty(self) -> bool:
        """Check if the current batch is empty.

        :return: True if current batch is empty, False otherwise.
        :rtype: bool
        """
        return len(self._batch) == 0

    def next(self) -> Any:
        """Pop the next item from the current batch.

        If current batch is empty/depleted, an API request is automatically
        sent to ArangoDB server to fetch the next batch and update the cursor.

        :return: Next item in current batch.
        :raise StopIteration: If the result set is depleted.
        :raise arango.exceptions.CursorNextError: If batch retrieval fails.
        :raise arango.exceptions.CursorStateError: If cursor ID is not set.
        """
        if self.empty():
            if not self.has_more():
                raise StopIteration
            self.fetch()

        return self.pop()

    def pop(self) -> Any:
        """Pop the next item from current batch.

        If current batch is empty/depleted, an exception is raised. You must
        call :func:`arango.cursor.Cursor.fetch` to manually fetch the next
        batch from server.

        :return: Next item in current batch.
        :raise arango.exceptions.CursorEmptyError: If current batch is empty.
        """
        if len(self._batch) == 0:
            raise CursorEmptyError("current batch is empty")
        return self._batch.popleft()

    def fetch(self) -> Json:
        """Fetch the next batch from server and update the cursor.

        :return: New batch details.
        :rtype: dict
        :raise arango.exceptions.CursorNextError: If batch retrieval fails.
        :raise arango.exceptions.CursorStateError: If cursor ID is not set.
        """
        if self._id is None:
            raise CursorStateError("cursor ID not set")
        request = Request(method="put", endpoint=f"/_api/{self._type}/{self._id}")
        resp = self._conn.send_request(request)

        if not resp.is_success:
            raise CursorNextError(resp, request)

        return self._update(resp.body)

    def close(self, ignore_missing: bool = False) -> Optional[bool]:
        """Close the cursor and free any server resources tied to it.

        :param ignore_missing: Do not raise exception on missing cursors.
        :type ignore_missing: bool
        :return: True if cursor was closed successfully, False if cursor was
            missing on the server and **ignore_missing** was set to True, None
            if there are no cursors to close server-side (e.g. result set is
            smaller than the batch size).
        :rtype: bool | None
        :raise arango.exceptions.CursorCloseError: If operation fails.
        :raise arango.exceptions.CursorStateError: If cursor ID is not set.
        """
        if self._id is None:
            return None
        request = Request(method="delete", endpoint=f"/_api/{self._type}/{self._id}")
        resp = self._conn.send_request(request)
        if resp.is_success:
            return True
        if resp.status_code == 404 and ignore_missing:
            return False
        raise CursorCloseError(resp, request)



__all__ = [
    "suppress_warning",
    "get_col_name",
    "get_doc_id",
    "is_none_or_int",
    "is_none_or_str",
]

import logging
from contextlib import contextmanager
from typing import Any, Iterator, Sequence, Union

from arango.exceptions import DocumentParseError
from arango.typings import Json


@contextmanager
def suppress_warning(logger_name: str) -> Iterator[None]:
    """Suppress logger messages.

    :param logger_name: Full name of the logger.
    :type logger_name: str
    """
    logger = logging.getLogger(logger_name)
    original_log_level = logger.getEffectiveLevel()
    logger.setLevel(logging.CRITICAL)
    yield
    logger.setLevel(original_log_level)


def get_col_name(doc: Union[str, Json]) -> str:
    """Return the collection name from input.

    :param doc: Document ID or body with "_id" field.
    :type doc: str | dict
    :return: Collection name.
    :rtype: str
    :raise arango.exceptions.DocumentParseError: If document ID is missing.
    """
    try:
        doc_id: str = doc["_id"] if isinstance(doc, dict) else doc
    except KeyError:
        raise DocumentParseError('field "_id" required')
    else:
        return doc_id.split("/", 1)[0]


def get_doc_id(doc: Union[str, Json]) -> str:
    """Return the document ID from input.

    :param doc: Document ID or body with "_id" field.
    :type doc: str | dict
    :return: Document ID.
    :rtype: str
    :raise arango.exceptions.DocumentParseError: If document ID is missing.
    """
    try:
        doc_id: str = doc["_id"] if isinstance(doc, dict) else doc
    except KeyError:
        raise DocumentParseError('field "_id" required')
    else:
        return doc_id


def is_none_or_int(obj: Any) -> bool:
    """Check if obj is None or an integer.

    :param obj: Object to check.
    :type obj: object
    :return: True if object is None or an integer.
    :rtype: bool
    """
    return obj is None or (isinstance(obj, int) and obj >= 0)


def is_none_or_str(obj: Any) -> bool:
    """Check if obj is None or a string.

    :param obj: Object to check.
    :type obj: object
    :return: True if object is None or a string.
    :rtype: bool
    """
    return obj is None or isinstance(obj, str)


def get_batches(elements: Sequence[Json], batch_size: int) -> Iterator[Sequence[Json]]:
    """Generator to split a list in batches
        of (maximum) **batch_size** elements each.

    :param elements: The list of elements.
    :type elements: Sequence[Json]
    :param batch_size: Max number of elements per batch.
    :type batch_size: int
    """
    for index in range(0, len(elements), batch_size):
        yield elements[index : index + batch_size]



__all__ = ["ArangoClient"]

from json import dumps, loads
from typing import Any, Callable, Optional, Sequence, Union

from pkg_resources import get_distribution

from arango.connection import (
    BasicConnection,
    Connection,
    JwtConnection,
    JwtSuperuserConnection,
)
from arango.database import StandardDatabase
from arango.exceptions import ServerConnectionError
from arango.http import DefaultHTTPClient, HTTPClient
from arango.resolver import (
    HostResolver,
    RandomHostResolver,
    RoundRobinHostResolver,
    SingleHostResolver,
)


class ArangoClient:
    """ArangoDB client.

    :param hosts: Host URL or list of URLs (coordinators in a cluster).
    :type hosts: str | [str]
    :param host_resolver: Host resolver. This parameter used for clusters (when
        multiple host URLs are provided). Accepted values are "roundrobin" and
        "random". Any other value defaults to round robin.
    :type host_resolver: str
    :param resolver_max_tries: Number of attempts to process an HTTP request
        before throwing a ConnectionAbortedError. Must not be lower than the
        number of hosts.
    :type resolver_max_tries: int
    :param http_client: User-defined HTTP client.
    :type http_client: arango.http.HTTPClient
    :param serializer: User-defined JSON serializer. Must be a callable
        which takes a JSON data type object as its only argument and return
        the serialized string. If not given, ``json.dumps`` is used by default.
    :type serializer: callable
    :param deserializer: User-defined JSON de-serializer. Must be a callable
        which takes a JSON serialized string as its only argument and return
        the de-serialized object. If not given, ``json.loads`` is used by
        default.
    :type deserializer: callable
    :param verify_override: Override TLS certificate verification. This will
       override the verify method of the underlying HTTP client.
       None: Do not change the verification behavior of the underlying HTTP client.
       True: Verify TLS certificate using the system CA certificates.
       False: Do not verify TLS certificate.
       str: Path to a custom CA bundle file or directory.
    :type verify_override: Union[bool, str, None]
    :param request_timeout: This is the default request timeout (in seconds)
       for http requests issued by the client if the parameter http_client is
       not secified. The default value is 60.
       None: No timeout.
       int: Timeout value in seconds.
    :type request_timeout: Any
    """

    def __init__(
        self,
        hosts: Union[str, Sequence[str]] = "http://127.0.0.1:8529",
        host_resolver: str = "roundrobin",
        resolver_max_tries: Optional[int] = None,
        http_client: Optional[HTTPClient] = None,
        serializer: Callable[..., str] = lambda x: dumps(x),
        deserializer: Callable[[str], Any] = lambda x: loads(x),
        verify_override: Union[bool, str, None] = None,
        request_timeout: Any = 60,
    ) -> None:
        if isinstance(hosts, str):
            self._hosts = [host.strip("/") for host in hosts.split(",")]
        else:
            self._hosts = [host.strip("/") for host in hosts]

        host_count = len(self._hosts)
        self._host_resolver: HostResolver

        if host_count == 1:
            self._host_resolver = SingleHostResolver(1, resolver_max_tries)
        elif host_resolver == "random":
            self._host_resolver = RandomHostResolver(host_count, resolver_max_tries)
        else:
            self._host_resolver = RoundRobinHostResolver(host_count, resolver_max_tries)

        # Initializes the http client
        self._http = http_client or DefaultHTTPClient()
        # Sets the request timeout.
        # This call can only happen AFTER initializing the http client.
        if http_client is None:
            self.request_timeout = request_timeout

        self._serializer = serializer
        self._deserializer = deserializer
        self._sessions = [self._http.create_session(h) for h in self._hosts]

        # override SSL/TLS certificate verification if provided
        if verify_override is not None:
            for session in self._sessions:
                session.verify = verify_override

    def __repr__(self) -> str:
        return f"<ArangoClient {','.join(self._hosts)}>"

    def close(self) -> None:  # pragma: no cover
        """Close HTTP sessions."""
        for session in self._sessions:
            session.close()

    @property
    def hosts(self) -> Sequence[str]:
        """Return the list of ArangoDB host URLs.

        :return: List of ArangoDB host URLs.
        :rtype: [str]
        """
        return self._hosts

    @property
    def version(self) -> str:
        """Return the client version.

        :return: Client version.
        :rtype: str
        """
        version: str = get_distribution("python-arango").version
        return version

    @property
    def request_timeout(self) -> Any:
        """Return the request timeout of the http client.

        :return: Request timeout.
        :rtype: Any
        """
        return self._http.REQUEST_TIMEOUT  # type: ignore

    # Setter for request_timeout
    @request_timeout.setter
    def request_timeout(self, value: Any) -> None:
        self._http.REQUEST_TIMEOUT = value  # type: ignore

    def db(
        self,
        name: str = "_system",
        username: str = "root",
        password: str = "",
        verify: bool = False,
        auth_method: str = "basic",
        superuser_token: Optional[str] = None,
        verify_certificate: bool = True,
    ) -> StandardDatabase:
        """Connect to an ArangoDB database and return the database API wrapper.

        :param name: Database name.
        :type name: str
        :param username: Username for basic authentication.
        :type username: str
        :param password: Password for basic authentication.
        :type password: str
        :param verify: Verify the connection by sending a test request.
        :type verify: bool
        :param auth_method: HTTP authentication method. Accepted values are
            "basic" (default) and "jwt". If set to "jwt", the token is
            refreshed automatically using ArangoDB username and password. This
            assumes that the clocks of the server and client are synchronized.
        :type auth_method: str
        :param superuser_token: User generated token for superuser access.
            If set, parameters **username**, **password** and **auth_method**
            are ignored. This token is not refreshed automatically.
        :type superuser_token: str
        :param verify_certificate: Verify TLS certificates.
        :type verify_certificate: bool
        :return: Standard database API wrapper.
        :rtype: arango.database.StandardDatabase
        :raise arango.exceptions.ServerConnectionError: If **verify** was set
            to True and the connection fails.
        """
        connection: Connection

        if superuser_token is not None:
            connection = JwtSuperuserConnection(
                hosts=self._hosts,
                host_resolver=self._host_resolver,
                sessions=self._sessions,
                db_name=name,
                http_client=self._http,
                serializer=self._serializer,
                deserializer=self._deserializer,
                superuser_token=superuser_token,
            )
        elif auth_method.lower() == "basic":
            connection = BasicConnection(
                hosts=self._hosts,
                host_resolver=self._host_resolver,
                sessions=self._sessions,
                db_name=name,
                username=username,
                password=password,
                http_client=self._http,
                serializer=self._serializer,
                deserializer=self._deserializer,
            )
        elif auth_method.lower() == "jwt":
            connection = JwtConnection(
                hosts=self._hosts,
                host_resolver=self._host_resolver,
                sessions=self._sessions,
                db_name=name,
                username=username,
                password=password,
                http_client=self._http,
                serializer=self._serializer,
                deserializer=self._deserializer,
            )
        else:
            raise ValueError(f"invalid auth_method: {auth_method}")

        if verify:
            try:
                connection.ping()
            except ServerConnectionError as err:
                raise err
            except Exception as err:
                raise ServerConnectionError(f"bad connection: {err}")

        return StandardDatabase(connection)



from typing import Any

from arango.typings import Headers, Json


def verify_format(_: Any, res: Json) -> Json:
    return res


def format_body(body: Json) -> Json:
    """Format generic response body.

    :param body: Input body.
    :type body: dict
    :return: Formatted body.
    :rtype: dict
    """
    body.pop("error", None)
    body.pop("code", None)
    return body


def format_index(body: Json) -> Json:
    """Format index data.

    :param body: Input body.
    :type body: dict
    :return: Formatted body.
    :rtype: dict
    """
    result = {"id": body["id"].split("/", 1)[-1], "fields": body["fields"]}
    if "type" in body:
        result["type"] = body["type"]
    if "name" in body:
        result["name"] = body["name"]
    if "deduplicate" in body:
        result["deduplicate"] = body["deduplicate"]
    if "sparse" in body:
        result["sparse"] = body["sparse"]
    if "unique" in body:
        result["unique"] = body["unique"]
    if "minLength" in body:
        result["min_length"] = body["minLength"]
    if "geoJson" in body:
        result["geo_json"] = body["geoJson"]
    if "ignoreNull" in body:
        result["ignore_none"] = body["ignoreNull"]
    if "selectivityEstimate" in body:
        result["selectivity"] = body["selectivityEstimate"]
    if "isNewlyCreated" in body:
        result["new"] = body["isNewlyCreated"]
    if "expireAfter" in body:
        result["expiry_time"] = body["expireAfter"]
    if "inBackground" in body:
        result["in_background"] = body["inBackground"]
    if "bestIndexedLevel" in body:
        result["best_indexed_level"] = body["bestIndexedLevel"]
    if "worstIndexedLevel" in body:
        result["worst_indexed_level"] = body["worstIndexedLevel"]
    if "maxNumCoverCells" in body:
        result["max_num_cover_cells"] = body["maxNumCoverCells"]
    if "storedValues" in body:
        result["storedValues"] = body["storedValues"]
    if "cacheEnabled" in body:
        result["cacheEnabled"] = body["cacheEnabled"]
    if "legacyPolygons" in body:
        result["legacyPolygons"] = body["legacyPolygons"]

    return verify_format(body, result)


def format_key_options(body: Json) -> Json:
    """Format collection key options data.

    :param body: Input body.
    :type body: dict
    :return: Formatted body.
    :rtype: dict
    """
    result: Json = {}

    if "type" in body:
        result["key_generator"] = body["type"]
    if "increment" in body:
        result["key_increment"] = body["increment"]
    if "offset" in body:
        result["key_offset"] = body["offset"]
    if "allowUserKeys" in body:
        result["user_keys"] = body["allowUserKeys"]
    if "lastValue" in body:
        result["key_last_value"] = body["lastValue"]

    return verify_format(body, result)


def format_database(body: Json) -> Json:
    """Format databases info.

    :param body: Input body.
    :type body: dict
    :return: Formatted body.
    :rtype: dict
    """
    result: Json = {}

    if "id" in body:
        result["id"] = body["id"]
    if "name" in body:
        result["name"] = body["name"]
    if "path" in body:
        result["path"] = body["path"]
    if "system" in body:
        result["system"] = body["system"]
    if "isSystem" in body:
        result["system"] = body["isSystem"]

    # Cluster only
    if "sharding" in body:
        result["sharding"] = body["sharding"]
    if "replicationFactor" in body:
        result["replication_factor"] = body["replicationFactor"]
    if "writeConcern" in body:
        result["write_concern"] = body["writeConcern"]

    return verify_format(body, result)


def format_collection(body: Json) -> Json:
    """Format collection data.

    :param body: Input body.
    :type body: dict
    :return: Formatted body.
    :rtype: dict
    """
    result: Json = {}

    if "id" in body:
        result["id"] = body["id"]
    if "objectId" in body:
        result["object_id"] = body["objectId"]
    if "name" in body:
        result["name"] = body["name"]
    if "isSystem" in body:
        result["system"] = body["isSystem"]
    if "isSmart" in body:
        result["smart"] = body["isSmart"]
    if "type" in body:
        result["type"] = body["type"]
        result["edge"] = body["type"] == 3
    if "waitForSync" in body:
        result["sync"] = body["waitForSync"]

    if "status" in body:
        result["status"] = body["status"]
    if "statusString" in body:
        result["status_string"] = body["statusString"]
    if "globallyUniqueId" in body:
        result["global_id"] = body["globallyUniqueId"]
    if "cacheEnabled" in body:
        result["cache"] = body["cacheEnabled"]
    if "replicationFactor" in body:
        result["replication_factor"] = body["replicationFactor"]
    if "minReplicationFactor" in body:
        result["min_replication_factor"] = body["minReplicationFactor"]
    if "writeConcern" in body:
        result["write_concern"] = body["writeConcern"]

    # Cluster only
    if "shards" in body:
        result["shards"] = body["shards"]
    if "replicationFactor" in body:
        result["replication_factor"] = body["replicationFactor"]
    if "numberOfShards" in body:
        result["shard_count"] = body["numberOfShards"]
    if "shardKeys" in body:
        result["shard_fields"] = body["shardKeys"]
    if "distributeShardsLike" in body:
        result["shard_like"] = body["distributeShardsLike"]
    if "shardingStrategy" in body:
        result["sharding_strategy"] = body["shardingStrategy"]
    if "smartJoinAttribute" in body:
        result["smart_join_attribute"] = body["smartJoinAttribute"]

    # Key Generator
    if "keyOptions" in body:
        result["key_options"] = format_key_options(body["keyOptions"])

    # Replication only
    if "cid" in body:
        result["cid"] = body["cid"]
    if "version" in body:
        result["version"] = body["version"]
    if "allowUserKeys" in body:
        result["user_keys"] = body["allowUserKeys"]
    if "planId" in body:
        result["plan_id"] = body["planId"]
    if "deleted" in body:
        result["deleted"] = body["deleted"]

    # New in 3.7
    if "syncByRevision" in body:
        result["sync_by_revision"] = body["syncByRevision"]
    if "tempObjectId" in body:
        result["temp_object_id"] = body["tempObjectId"]
    if "usesRevisionsAsDocumentIds" in body:
        result["rev_as_id"] = body["usesRevisionsAsDocumentIds"]
    if "isDisjoint" in body:
        result["disjoint"] = body["isDisjoint"]
    if "isSmartChild" in body:
        result["smart_child"] = body["isSmartChild"]
    if "minRevision" in body:
        result["min_revision"] = body["minRevision"]
    if "schema" in body:
        result["schema"] = body["schema"]

    # New in 3.10
    if "computedValues" in body:
        result["computedValues"] = [
            {
                "name": cv["name"],
                "expression": cv["expression"],
                "overwrite": cv["overwrite"],
                "computedOn": cv["computedOn"],
                "keepNull": cv["keepNull"],
                "failOnWarning": cv["failOnWarning"],
            }
            for cv in body["computedValues"]
        ]

    return verify_format(body, result)


def format_aql_cache(body: Json) -> Json:
    """Format AQL cache data.

    :param body: Input body.
    :type body: dict
    :return: Formatted body.
    :rtype: dict
    """
    result = {
        "mode": body["mode"],
        "max_results": body["maxResults"],
        "max_results_size": body["maxResultsSize"],
        "max_entry_size": body["maxEntrySize"],
        "include_system": body["includeSystem"],
    }
    return verify_format(body, result)


def format_wal_properties(body: Json) -> Json:
    """Format WAL properties.

    :param body: Input body.
    :type body: dict
    :return: Formatted body.
    :rtype: dict
    """
    result: Json = {}
    if "allowOversizeEntries" in body:
        result["oversized_ops"] = body["allowOversizeEntries"]
    if "logfileSize" in body:
        result["log_size"] = body["logfileSize"]
    if "historicLogfiles" in body:
        result["historic_logs"] = body["historicLogfiles"]
    if "reserveLogfiles" in body:
        result["reserve_logs"] = body["reserveLogfiles"]
    if "syncInterval" in body:
        result["sync_interval"] = body["syncInterval"]
    if "throttleWait" in body:
        result["throttle_wait"] = body["throttleWait"]
    if "throttleWhenPending" in body:
        result["throttle_limit"] = body["throttleWhenPending"]

    return verify_format(body, result)


def format_wal_transactions(body: Json) -> Json:
    """Format WAL transactions.

    :param body: Input body.
    :type body: dict
    :return: Formatted body.
    :rtype: dict
    """
    result: Json = {}
    if "minLastCollected" in body:
        result["last_collected"] = body["minLastCollected"]
    if "minLastSealed" in body:
        result["last_sealed"] = body["minLastSealed"]
    if "runningTransactions" in body:
        result["count"] = body["runningTransactions"]

    return verify_format(body, result)


def format_aql_query(body: Json) -> Json:
    """Format AQL query data.

    :param body: Input body.
    :type body: dict
    :return: Formatted body.
    :rtype: dict
    """
    result = {"id": body["id"], "query": body["query"]}
    if "database" in body:
        result["database"] = body["database"]
    if "bindVars" in body:
        result["bind_vars"] = body["bindVars"]
    if "runTime" in body:
        result["runtime"] = body["runTime"]
    if "started" in body:
        result["started"] = body["started"]
    if "state" in body:
        result["state"] = body["state"]
    if "stream" in body:
        result["stream"] = body["stream"]
    if "user" in body:
        result["user"] = body["user"]
    return verify_format(body, result)


def format_aql_tracking(body: Json) -> Json:
    """Format AQL tracking data.

    :param body: Input body.
    :type body: dict
    :return: Formatted body.
    :rtype: dict
    """
    result: Json = {}
    if "enabled" in body:
        result["enabled"] = body["enabled"]
    if "maxQueryStringLength" in body:
        result["max_query_string_length"] = body["maxQueryStringLength"]
    if "maxSlowQueries" in body:
        result["max_slow_queries"] = body["maxSlowQueries"]
    if "slowQueryThreshold" in body:
        result["slow_query_threshold"] = body["slowQueryThreshold"]
    if "slowStreamingQueryThreshold" in body:
        result["slow_streaming_query_threshold"] = body["slowStreamingQueryThreshold"]
    if "trackBindVars" in body:
        result["track_bind_vars"] = body["trackBindVars"]
    if "trackSlowQueries" in body:
        result["track_slow_queries"] = body["trackSlowQueries"]

    return verify_format(body, result)


def format_tick_values(body: Json) -> Json:
    """Format tick data.

    :param body: Input body.
    :type body: dict
    :return: Formatted body.
    :rtype: dict
    """
    result: Json = {}

    if "tickMin" in body:
        result["tick_min"] = body["tickMin"]
    if "tickMax" in body:
        result["tick_max"] = body["tickMax"]
    if "tick" in body:
        result["tick"] = body["tick"]
    if "time" in body:
        result["time"] = body["time"]
    if "server" in body:
        result["server"] = format_server_info(body["server"])

    return verify_format(body, result)


def format_server_info(body: Json) -> Json:
    """Format server data.

    :param body: Input body.
    :type body: dict
    :return: Formatted body.
    :rtype: dict
    """
    return {"version": body["version"], "server_id": body["serverId"]}


def format_server_status(body: Json) -> Json:
    """Format server status.

    :param body: Input body.
    :type body: dict
    :return: Formatted body.
    :rtype: dict
    """
    result: Json = {}

    if "foxxApi" in body:
        result["foxx_api"] = body["foxxApi"]
    if "host" in body:
        result["host"] = body["host"]
    if "hostname" in body:
        result["hostname"] = body["hostname"]
    if "license" in body:
        result["license"] = body["license"]
    if "mode" in body:
        result["mode"] = body["mode"]
    if "operationMode" in body:
        result["operation_mode"] = body["operationMode"]
    if "pid" in body:
        result["pid"] = body["pid"]
    if "server" in body:
        result["server"] = body["server"]
    if "serverInfo" in body:
        info = body["serverInfo"]
        if "writeOpsEnabled" in info:
            info["write_ops_enabled"] = info.pop("writeOpsEnabled")
        if "readOnly" in info:
            info["read_only"] = info.pop("readOnly")
        result["server_info"] = info
    if "version" in body:
        result["version"] = body["version"]

    return verify_format(body, result)


def format_replication_applier_config(body: Json) -> Json:
    """Format replication applier configuration data.

    :param body: Input body.
    :type body: dict
    :return: Formatted body.
    :rtype: dict
    """
    result: Json = {}
    if "endpoint" in body:
        result["endpoint"] = body["endpoint"]
    if "database" in body:
        result["database"] = body["database"]
    if "username" in body:
        result["username"] = body["username"]
    if "verbose" in body:
        result["verbose"] = body["verbose"]
    if "incremental" in body:
        result["incremental"] = body["incremental"]
    if "requestTimeout" in body:
        result["request_timeout"] = body["requestTimeout"]
    if "connectTimeout" in body:
        result["connect_timeout"] = body["connectTimeout"]
    if "ignoreErrors" in body:
        result["ignore_errors"] = body["ignoreErrors"]
    if "maxConnectRetries" in body:
        result["max_connect_retries"] = body["maxConnectRetries"]
    if "lockTimeoutRetries" in body:
        result["lock_timeout_retries"] = body["lockTimeoutRetries"]
    if "sslProtocol" in body:
        result["ssl_protocol"] = body["sslProtocol"]
    if "chunkSize" in body:
        result["chunk_size"] = body["chunkSize"]
    if "skipCreateDrop" in body:
        result["skip_create_drop"] = body["skipCreateDrop"]
    if "autoStart" in body:
        result["auto_start"] = body["autoStart"]
    if "adaptivePolling" in body:
        result["adaptive_polling"] = body["adaptivePolling"]
    if "autoResync" in body:
        result["auto_resync"] = body["autoResync"]
    if "autoResyncRetries" in body:
        result["auto_resync_retries"] = body["autoResyncRetries"]
    if "maxPacketSize" in body:
        result["max_packet_size"] = body["maxPacketSize"]
    if "includeSystem" in body:
        result["include_system"] = body["includeSystem"]
    if "includeFoxxQueues" in body:
        result["include_foxx_queues"] = body["includeFoxxQueues"]
    if "requireFromPresent" in body:
        result["require_from_present"] = body["requireFromPresent"]
    if "restrictType" in body:
        result["restrict_type"] = body["restrictType"]
    if "restrictCollections" in body:
        result["restrict_collections"] = body["restrictCollections"]
    if "connectionRetryWaitTime" in body:
        result["connection_retry_wait_time"] = body["connectionRetryWaitTime"]
    if "initialSyncMaxWaitTime" in body:
        result["initial_sync_max_wait_time"] = body["initialSyncMaxWaitTime"]
    if "idleMinWaitTime" in body:
        result["idle_min_wait_time"] = body["idleMinWaitTime"]
    if "idleMaxWaitTime" in body:
        result["idle_max_wait_time"] = body["idleMaxWaitTime"]

    return verify_format(body, result)


def format_applier_progress(body: Json) -> Json:
    """Format replication applier progress data.

    :param body: Input body.
    :type body: dict
    :return: Formatted body.
    :rtype: dict
    """
    result: Json = {}
    if "time" in body:
        result["time"] = body["time"]
    if "message" in body:
        result["message"] = body["message"]
    if "failedConnects" in body:
        result["failed_connects"] = body["failedConnects"]

    return verify_format(body, result)


def format_applier_error(body: Json) -> Json:
    """Format replication applier error data.

    :param body: Input body.
    :type body: dict
    :return: Formatted body.
    :rtype: dict
    """
    result: Json = {}
    if "errorNum" in body:
        result["error_num"] = body["errorNum"]
    if "errorMessage" in body:
        result["error_message"] = body["errorMessage"]
    if "time" in body:
        result["time"] = body["time"]

    return verify_format(body, result)


def format_applier_state_details(body: Json) -> Json:
    """Format replication applier state details.

    :param body: Input body.
    :type body: dict
    :return: Formatted body.
    :rtype: dict
    """
    result: Json = {}
    if "started" in body:
        result["started"] = body["started"]
    if "running" in body:
        result["running"] = body["running"]
    if "phase" in body:
        result["phase"] = body["phase"]
    if "time" in body:
        result["time"] = body["time"]
    if "safeResumeTick" in body:
        result["safe_resume_tick"] = body["safeResumeTick"]
    if "ticksBehind" in body:
        result["ticks_behind"] = body["ticksBehind"]
    if "lastAppliedContinuousTick" in body:
        result["last_applied_continuous_tick"] = body["lastAppliedContinuousTick"]
    if "lastProcessedContinuousTick" in body:
        result["last_processed_continuous_tick"] = body["lastProcessedContinuousTick"]
    if "lastAvailableContinuousTick" in body:
        result["last_available_continuous_tick"] = body["lastAvailableContinuousTick"]
    if "progress" in body:
        result["progress"] = format_applier_progress(body["progress"])
    if "totalRequests" in body:
        result["total_requests"] = body["totalRequests"]
    if "totalFailedConnects" in body:
        result["total_failed_connects"] = body["totalFailedConnects"]
    if "totalEvents" in body:
        result["total_events"] = body["totalEvents"]
    if "totalDocuments" in body:
        result["total_documents"] = body["totalDocuments"]
    if "totalRemovals" in body:
        result["total_removals"] = body["totalRemovals"]
    if "totalResyncs" in body:
        result["total_resyncs"] = body["totalResyncs"]
    if "totalOperationsExcluded" in body:
        result["total_operations_excluded"] = body["totalOperationsExcluded"]
    if "totalApplyTime" in body:
        result["total_apply_time"] = body["totalApplyTime"]
    if "averageApplyTime" in body:
        result["average_apply_time"] = body["averageApplyTime"]
    if "totalFetchTime" in body:
        result["total_fetch_time"] = body["totalFetchTime"]
    if "averageFetchTime" in body:
        result["average_fetch_time"] = body["averageFetchTime"]
    if "lastError" in body:
        result["last_error"] = format_applier_error(body["lastError"])

    return verify_format(body, result)


def format_replication_applier_state(body: Json) -> Json:
    """Format replication applier state.

    :param body: Input body.
    :type body: dict
    :return: Formatted body.
    :rtype: dict
    """
    result: Json = {}
    if "endpoint" in body:
        result["endpoint"] = body["endpoint"]
    if "database" in body:
        result["database"] = body["database"]
    if "username" in body:
        result["username"] = body["username"]
    if "state" in body:
        result["state"] = format_applier_state_details(body["state"])
    if "server" in body:
        result["server"] = format_server_info(body["server"])

    return verify_format(body, result)


def format_replication_state(body: Json) -> Json:
    """Format replication state.

    :param body: Input body.
    :type body: dict
    :return: Formatted body.
    :rtype: dict
    """
    if not isinstance(body, dict):
        return body

    result: Json = {}
    if "running" in body:
        result["running"] = body["running"]
    if "time" in body:
        result["time"] = body["time"]
    if "lastLogTick" in body:
        result["last_log_tick"] = body["lastLogTick"]
    if "totalEvents" in body:
        result["total_events"] = body["totalEvents"]
    if "lastUncommittedLogTick" in body:
        result["last_uncommitted_log_tick"] = body["lastUncommittedLogTick"]

    return verify_format(body, result)


def format_replication_logger_state(body: Json) -> Json:
    """Format replication collection data.

    :param body: Input body.
    :type body: dict
    :return: Formatted body.
    :rtype: dict
    """
    result: Json = {}
    if "state" in body:
        result["state"] = format_replication_state(body["state"])
    if "server" in body:
        result["server"] = format_server_info(body["server"])
    if "clients" in body:
        result["clients"] = body["clients"]

    return verify_format(body, result)


def format_replication_collection(body: Json) -> Json:
    """Format replication collection data.

    :param body: Input body.
    :type body: dict
    :return: Formatted body.
    :rtype: dict
    """
    result: Json = {}
    if "planVersion" in body:
        result["plan_version"] = body["planVersion"]
    if "isReady" in body:
        result["is_ready"] = body["isReady"]
    if "allInSync" in body:
        result["all_in_sync"] = body["allInSync"]
    if "indexes" in body:
        result["indexes"] = [format_index(index) for index in body["indexes"]]
    if "parameters" in body:
        result["parameters"] = format_collection(body["parameters"])

    return verify_format(body, result)


def format_replication_database(body: Json) -> Json:
    """Format replication database data.

    :param body: Input body.
    :type body: dict
    :return: Formatted body.
    :rtype: dict
    """
    result = {
        "id": body["id"],
        "name": body["name"],
        "collections": [
            format_replication_collection(col) for col in body["collections"]
        ],
        "views": [format_view(view) for view in body["views"]],
    }
    if "properties" in body:
        result["properties"] = format_database(body["properties"])

    return verify_format(body, result)


def format_replication_inventory(body: Json) -> Json:
    """Format replication inventory data.

    :param body: Input body.
    :type body: dict
    :return: Formatted body.
    :rtype: dict
    """
    result: Json = {}
    if "tick" in body:
        result["tick"] = body["tick"]
    if "state" in body:
        result["state"] = format_replication_state(body["state"])

    if "databases" in body:
        result["databases"] = {
            k: format_replication_database(v) for k, v in body["databases"].items()
        }
    if "collections" in body:
        result["collections"] = [
            format_replication_collection(col) for col in body["collections"]
        ]
    if "views" in body:
        result["views"] = [format_view(view) for view in body["views"]]
    if "properties" in body:
        result["properties"] = format_database(body["properties"])

    return verify_format(body, result)


def format_replication_sync(body: Json) -> Json:
    """Format replication sync result.

    :param body: Input body.
    :type body: dict
    :return: Formatted body.
    :rtype: dict
    """
    result: Json = {}
    if "collections" in body:
        result["collections"] = body["collections"]
    if "lastLogTick" in body:
        result["last_log_tick"] = body["lastLogTick"]
    return verify_format(body, result)


def format_replication_header(headers: Headers) -> Json:
    """Format replication headers.

    :param headers: Request headers.
    :type headers: dict
    :return: Formatted body.
    :rtype: dict
    """
    headers = {k.lower(): v for k, v in headers.items()}
    result: Json = {}

    if "x-arango-replication-frompresent" in headers:
        result["from_present"] = headers["x-arango-replication-frompresent"] == "true"

    if "x-arango-replication-lastincluded" in headers:
        result["last_included"] = headers["x-arango-replication-lastincluded"]

    if "x-arango-replication-lastscanned" in headers:
        result["last_scanned"] = headers["x-arango-replication-lastscanned"]

    if "x-arango-replication-lasttick" in headers:
        result["last_tick"] = headers["x-arango-replication-lasttick"]

    if "x-arango-replication-active" in headers:
        result["active"] = headers["x-arango-replication-active"] == "true"

    if "x-arango-replication-checkmore" in headers:
        result["check_more"] = headers["x-arango-replication-checkmore"] == "true"

    return result


def format_view_link(body: Json) -> Json:
    """Format view link data.

    :param body: Input body.
    :type body: dict
    :return: Formatted body.
    :rtype: dict
    """
    result: Json = {}
    if "analyzers" in body:
        result["analyzers"] = body["analyzers"]
    if "fields" in body:
        result["fields"] = body["fields"]
    if "includeAllFields" in body:
        result["include_all_fields"] = body["includeAllFields"]
    if "trackListPositions" in body:
        result["track_list_positions"] = body["trackListPositions"]
    if "storeValues" in body:
        result["store_values"] = body["storeValues"]

    return verify_format(body, result)


def format_view_consolidation_policy(body: Json) -> Json:
    """Format view consolidation policy data.

    :param body: Input body.
    :type body: dict
    :return: Formatted body.
    :rtype: dict
    """
    result: Json = {}
    if "type" in body:
        result["type"] = body["type"]
    if "threshold" in body:
        result["threshold"] = body["threshold"]
    if "segmentsMin" in body:
        result["segments_min"] = body["segmentsMin"]
    if "segmentsMax" in body:
        result["segments_max"] = body["segmentsMax"]
    if "segmentsBytesMax" in body:
        result["segments_bytes_max"] = body["segmentsBytesMax"]
    if "segmentsBytesFloor" in body:
        result["segments_bytes_floor"] = body["segmentsBytesFloor"]
    if "minScore" in body:
        result["min_score"] = body["minScore"]

    return verify_format(body, result)


def format_view(body: Json) -> Json:
    """Format view data.

    :param body: Input body.
    :type body: dict
    :return: Formatted body.
    :rtype: dict
    """
    result: Json = {}
    if "globallyUniqueId" in body:
        result["global_id"] = body["globallyUniqueId"]
    if "id" in body:
        result["id"] = body["id"]
    if "name" in body:
        result["name"] = body["name"]
    if "type" in body:
        result["type"] = body["type"]
    if "cleanupIntervalStep" in body:
        result["cleanup_interval_step"] = body["cleanupIntervalStep"]
    if "commitIntervalMsec" in body:
        result["commit_interval_msec"] = body["commitIntervalMsec"]
    if "consolidationIntervalMsec" in body:
        result["consolidation_interval_msec"] = body["consolidationIntervalMsec"]
    if "consolidationPolicy" in body:
        result["consolidation_policy"] = format_view_consolidation_policy(
            body["consolidationPolicy"]
        )
    if "primarySort" in body:
        result["primary_sort"] = body["primarySort"]
    if "primarySortCompression" in body:
        result["primary_sort_compression"] = body["primarySortCompression"]
    if "storedValues" in body:
        result["stored_values"] = body["storedValues"]
    if "writebufferIdle" in body:
        result["writebuffer_idle"] = body["writebufferIdle"]
    if "writebufferActive" in body:
        result["writebuffer_active"] = body["writebufferActive"]
    if "writebufferSizeMax" in body:
        result["writebuffer_max_size"] = body["writebufferSizeMax"]
    if "links" in body:
        result["links"] = {
            name: format_view_link(link) for name, link in body["links"].items()
        }

    return verify_format(body, result)


def format_vertex(body: Json) -> Json:
    """Format vertex data.

    :param body: Input body.
    :type body: dict
    :return: Formatted body.
    :rtype: dict
    """
    vertex: Json = body["vertex"]
    if "_oldRev" in vertex:
        vertex["_old_rev"] = vertex.pop("_oldRev")

    if "new" in body or "old" in body:
        result: Json = {"vertex": vertex}
        if "new" in body:
            result["new"] = body["new"]
        if "old" in body:
            result["old"] = body["old"]
        return result
    else:
        return vertex


def format_edge(body: Json) -> Json:
    """Format edge data.

    :param body: Input body.
    :type body: dict
    :return: Formatted body.
    :rtype: dict
    """
    edge: Json = body["edge"]
    if "_oldRev" in edge:
        edge["_old_rev"] = edge.pop("_oldRev")

    if "new" in body or "old" in body:
        result: Json = {"edge": edge}
        if "new" in body:
            result["new"] = body["new"]
        if "old" in body:
            result["old"] = body["old"]
        return result
    else:
        return edge


def format_tls(body: Json) -> Json:
    """Format TLS data.

    :param body: Input body.
    :type body: dict
    :return: Formatted body.
    :rtype: dict
    """
    result = body
    return verify_format(body, result)


def format_backup(body: Json) -> Json:
    """Format backup entry.

    :param body: Input body.
    :type body: dict
    :return: Formatted body.
    :rtype: dict
    """
    result: Json = {}

    if "previous" in body:
        result["previous"] = body["previous"]
    if "id" in body:
        result["backup_id"] = body["id"]
    if "datetime" in body:
        result["datetime"] = body["datetime"]
    if "potentiallyInconsistent" in body:
        result["potentially_inconsistent"] = body["potentiallyInconsistent"]
    if "sizeInBytes" in body:
        result["size_in_bytes"] = body["sizeInBytes"]
    if "nrDBServers" in body:
        result["dbserver_count"] = body["nrDBServers"]
    if "nrFiles" in body:
        result["file_count"] = body["nrFiles"]

    if "available" in body:
        result["available"] = body["available"]
    if "version" in body:
        result["version"] = body["version"]
    if "keys" in body:
        result["keys"] = body["keys"]
    if "nrPiecesPresent" in body:
        result["pieces_present"] = body["nrPiecesPresent"]

    return verify_format(body, result)


def format_backups(body: Json) -> Json:
    """Format backup entries.

    :param body: Input body.
    :type body: dict
    :return: Formatted body.
    :rtype: dict
    """
    result: Json = {}

    if "server" in body:
        result["server"] = body["server"]
    if "list" in body:
        result["list"] = {
            key: format_backup(backup) for key, backup in body["list"].items()
        }
    return verify_format(body, result)


def format_backup_restore(body: Json) -> Json:
    """Format backup restore data.

    :param body: Input body.
    :type body: dict
    :return: Formatted body.
    :rtype: dict
    """
    result: Json = {}
    if "id" in body:
        result["backup_id"] = body["id"]
    if "isCluster" in body:
        result["is_cluster"] = body["isCluster"]
    if "previous" in body:
        result["previous"] = body["previous"]

    return verify_format(body, result)


def format_backup_dbserver(body: Json) -> Json:
    """Format backup DBserver data.

    :param body: Input body.
    :type body: dict
    :return: Formatted body.
    :rtype: dict
    """
    return {"status": body["Status"]}


def format_backup_transfer(body: Json) -> Json:
    """Format backup download/upload data.

    :param body: Input body.
    :type body: dict
    :return: Formatted body.
    :rtype: dict
    """
    result: Json = {}
    if "Timestamp" in body:
        result["timestamp"] = body["Timestamp"]
    if "DownloadId" in body:
        result["download_id"] = body["DownloadId"]
    if "downloadId" in body:
        result["download_id"] = body["downloadId"]
    if "UploadId" in body:
        result["upload_id"] = body["UploadId"]
    if "uploadId" in body:
        result["upload_id"] = body["uploadId"]
    if "Cancelled" in body:
        result["cancelled"] = body["Cancelled"]
    if "BackupId" in body:
        result["backup_id"] = body["BackupId"]
    if "DBServers" in body:
        result["dbservers"] = {
            k: format_backup_dbserver(v) for k, v in body["DBServers"].items()
        }
    return verify_format(body, result)


def format_service_data(body: Json) -> Json:
    """Format Foxx service data.

    :param body: Input body.
    :type body: dict
    :return: Formatted body.
    :rtype: dict
    """
    if "manifest" in body:
        manifest = body["manifest"]
        if "defaultDocument" in manifest:
            manifest["default_document"] = manifest.pop("defaultDocument")

    return body


def format_pregel_job_data(body: Json) -> Json:
    """Format Pregel job data.

    :param body: Input body.
    :type body: dict
    :return: Formatted body.
    :rtype: dict
    """
    result: Json = {}

    if "aggregators" in body:
        result["aggregators"] = body["aggregators"]
    if "computationTime" in body:
        result["computation_time"] = body["computationTime"]
    if "edgeCount" in body:
        result["edge_count"] = body["edgeCount"]
    if "gss" in body:
        result["gss"] = body["gss"]
    if "receivedCount" in body:
        result["received_count"] = body["receivedCount"]
    if "sendCount" in body:
        result["send_count"] = body["sendCount"]
    if "startupTime" in body:
        result["startup_time"] = body["startupTime"]
    if "state" in body:
        result["state"] = body["state"]
    if "totalRuntime" in body:
        result["total_runtime"] = body["totalRuntime"]
    if "vertexCount" in body:
        result["vertex_count"] = body["vertexCount"]

    return verify_format(body, result)


def format_graph_properties(body: Json) -> Json:
    """Format graph properties.

    :param body: Input body.
    :type body: dict
    :return: Formatted body.
    :rtype: dict
    """
    result = {
        "id": body["_id"],
        "key": body["_key"],
        "name": body["name"],
        "revision": body["_rev"],
        "orphan_collections": body["orphanCollections"],
        "edge_definitions": [
            {
                "edge_collection": edge_definition["collection"],
                "from_vertex_collections": edge_definition["from"],
                "to_vertex_collections": edge_definition["to"],
            }
            for edge_definition in body["edgeDefinitions"]
        ],
    }
    if "isSmart" in body:
        result["smart"] = body["isSmart"]
    if "smartGraphAttribute" in body:
        result["smart_field"] = body["smartGraphAttribute"]
    if "numberOfShards" in body:
        result["shard_count"] = body["numberOfShards"]
    if "replicationFactor" in body:
        result["replication_factor"] = body["replicationFactor"]

    return verify_format(body, result)


def format_query_cache_entry(body: Json) -> Json:
    """Format AQL query cache entry.

    :param body: Input body.
    :type body: dict
    :return: Formatted body.
    :rtype: dict
    """
    result = {}

    if "hash" in body:
        result["hash"] = body["hash"]
    if "query" in body:
        result["query"] = body["query"]
    if "bindVars" in body:
        result["bind_vars"] = body["bindVars"]
    if "size" in body:
        result["size"] = body["size"]
    if "results" in body:
        result["results"] = body["results"]
    if "started" in body:
        result["started"] = body["started"]
    if "hits" in body:
        result["hits"] = body["hits"]
    if "runTime" in body:
        result["runtime"] = body["runTime"]
    if "dataSources" in body:
        result["data_sources"] = body["dataSources"]

    return verify_format(body, result)


def format_query_rule_item(body: Json) -> Json:
    """Format AQL query rule item.

    :param body: Input body.
    :type body: dict
    :return: Formatted body.
    :rtype: dict
    """
    result = {}

    if "name" in body:
        result["name"] = body["name"]
    if "flags" in body:
        result["flags"] = format_query_rule_item_flags(body["flags"])

    return verify_format(body, result)


def format_query_rule_item_flags(body: Json) -> Json:
    """Format AQL query rule item flags.

    :param body: Input body.
    :type body: dict
    :return: Formatted body.
    :rtype: dict
    """
    result = {}

    if "hidden" in body:
        result["hidden"] = body["hidden"]
    if "clusterOnly" in body:
        result["clusterOnly"] = body["clusterOnly"]
    if "canBeDisabled" in body:
        result["canBeDisabled"] = body["canBeDisabled"]
    if "canCreateAdditionalPlans" in body:
        result["canCreateAdditionalPlans"] = body["canCreateAdditionalPlans"]
    if "disabledByDefault" in body:
        result["disabledByDefault"] = body["disabledByDefault"]
    if "enterpriseOnly" in body:
        result["enterpriseOnly"] = body["enterpriseOnly"]

    return verify_format(body, result)



__all__ = ["Cluster"]

from typing import List

from arango.api import ApiGroup
from arango.exceptions import (
    ClusterEndpointsError,
    ClusterHealthError,
    ClusterMaintenanceModeError,
    ClusterServerCountError,
    ClusterServerEngineError,
    ClusterServerIDError,
    ClusterServerRoleError,
    ClusterServerStatisticsError,
    ClusterServerVersionError,
)
from arango.formatter import format_body
from arango.request import Request
from arango.response import Response
from arango.result import Result
from arango.typings import Json


class Cluster(ApiGroup):  # pragma: no cover
    def server_id(self) -> Result[str]:
        """Return the server ID.

        :return: Server ID.
        :rtype: str
        :raise arango.exceptions.ClusterServerIDError: If retrieval fails.
        """
        request = Request(method="get", endpoint="/_admin/server/id")

        def response_handler(resp: Response) -> str:
            if resp.is_success:
                return str(resp.body["id"])
            raise ClusterServerIDError(resp, request)

        return self._execute(request, response_handler)

    def server_role(self) -> Result[str]:
        """Return the server role.

        :return: Server role. Possible values are "SINGLE" (server which is
            not in a cluster), "COORDINATOR" (cluster coordinator), "PRIMARY",
            "SECONDARY", "AGENT" (Agency server in a cluster) or "UNDEFINED".
        :rtype: str
        :raise arango.exceptions.ClusterServerRoleError: If retrieval fails.
        """
        request = Request(method="get", endpoint="/_admin/server/role")

        def response_handler(resp: Response) -> str:
            if resp.is_success:
                return str(resp.body["role"])
            raise ClusterServerRoleError(resp, request)

        return self._execute(request, response_handler)

    def server_version(self, server_id: str) -> Result[Json]:
        """Return the version of the given server.

        :param server_id: Server ID.
        :type server_id: str
        :return: Version of the given server.
        :rtype: dict
        :raise arango.exceptions.ClusterServerVersionError: If retrieval fails.
        """
        request = Request(
            method="get",
            endpoint="/_admin/cluster/nodeVersion",
            params={"ServerID": server_id},
        )

        def response_handler(resp: Response) -> Json:
            if resp.is_success:
                return format_body(resp.body)
            raise ClusterServerVersionError(resp, request)

        return self._execute(request, response_handler)

    def server_engine(self, server_id: str) -> Result[Json]:
        """Return the engine details for the given server.

        :param server_id: Server ID.
        :type server_id: str
        :return: Engine details of the given server.
        :rtype: dict
        :raise arango.exceptions.ClusterServerEngineError: If retrieval fails.
        """
        request = Request(
            method="get",
            endpoint="/_admin/cluster/nodeEngine",
            params={"ServerID": server_id},
        )

        def response_handler(resp: Response) -> Json:
            if resp.is_success:
                return format_body(resp.body)
            raise ClusterServerEngineError(resp, request)

        return self._execute(request, response_handler)

    def server_count(self) -> Result[int]:
        """Return the number of servers in the cluster.

        :return: Number of servers in the cluster.
        :rtype: int
        :raise arango.exceptions.ClusterServerCountError: If retrieval fails.
        """
        request = Request(method="get", endpoint="/_admin/cluster/numberOfServers")

        def response_handler(resp: Response) -> int:
            if resp.is_success:
                result: int = resp.body
                return result
            raise ClusterServerCountError(resp, request)

        return self._execute(request, response_handler)

    def server_statistics(self, server_id: str) -> Result[Json]:
        """Return the statistics for the given server.

        :param server_id: Server ID.
        :type server_id: str
        :return: Statistics for the given server.
        :rtype: dict
        :raise arango.exceptions.ClusterServerStatisticsError: If retrieval fails.
        """
        request = Request(
            method="get",
            endpoint="/_admin/cluster/nodeStatistics",
            params={"ServerID": server_id},
        )

        def response_handler(resp: Response) -> Json:
            if resp.is_success:
                return format_body(resp.body)
            raise ClusterServerStatisticsError(resp, request)

        return self._execute(request, response_handler)

    def health(self) -> Result[Json]:
        """Return the cluster health.

        :return: Cluster health.
        :rtype: dict
        :raise arango.exceptions.ClusterHealthError: If retrieval fails.
        """
        request = Request(
            method="get",
            endpoint="/_admin/cluster/health",
        )

        def response_handler(resp: Response) -> Json:
            if resp.is_success:
                return format_body(resp.body)
            raise ClusterHealthError(resp, request)

        return self._execute(request, response_handler)

    def toggle_maintenance_mode(self, mode: str) -> Result[Json]:
        """Enable or disable the cluster supervision (agency) maintenance mode.

        :param mode: Maintenance mode. Allowed values are "on" and "off".
        :type mode: str
        :return: Result of the operation.
        :rtype: dict
        :raise arango.exceptions.ClusterMaintenanceModeError: If toggle fails.
        """
        request = Request(
            method="put",
            endpoint="/_admin/cluster/maintenance",
            data=f'"{mode}"',
        )

        def response_handler(resp: Response) -> Json:
            if resp.is_success:
                return format_body(resp.body)
            raise ClusterMaintenanceModeError(resp, request)

        return self._execute(request, response_handler)

    def endpoints(self) -> Result[List[str]]:
        """Return coordinate endpoints. This method is for clusters only.

        :return: List of endpoints.
        :rtype: [str]
        :raise arango.exceptions.ServerEndpointsError: If retrieval fails.
        """
        request = Request(method="get", endpoint="/_api/cluster/endpoints")

        def response_handler(resp: Response) -> List[str]:
            if not resp.is_success:
                raise ClusterEndpointsError(resp, request)
            return [item["endpoint"] for item in resp.body["endpoints"]]

        return self._execute(request, response_handler)



from setuptools import find_packages, setup

with open("./README.md") as fp:
    long_description = fp.read()

setup(
    name="python-arango",
    description="Python Driver for ArangoDB",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Joohwan Oh",
    author_email="joohwan.oh@outlook.com",
    url="https://github.com/ArangoDB-Community/python-arango",
    keywords=["arangodb", "python", "driver"],
    packages=find_packages(exclude=["tests"]),
    include_package_data=True,
    python_requires=">=3.7",
    license="MIT",
    install_requires=[
        "urllib3>=1.26.0",
        "dataclasses>=0.6; python_version < '3.7'",
        "requests",
        "requests_toolbelt",
        "PyJWT",
        "setuptools>=42",
    ],
    extras_require={
        "dev": [
            "black>=22.3.0",
            "flake8>=4.0.1",
            "isort>=5.10.1",
            "mypy>=0.942",
            "mock",
            "pre-commit>=2.17.0",
            "pytest>=7.1.1",
            "pytest-cov>=3.0.0",
            "sphinx",
            "sphinx_rtd_theme",
            "types-pkg_resources",
            "types-requests",
            "types-setuptools",
        ],
    },
    classifiers=[
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Operating System :: MacOS",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: Unix",
        "Programming Language :: Python :: 3",
        "Topic :: Documentation :: Sphinx",
    ],
)



__all__ = [
    "BaseConnection",
    "BasicConnection",
    "Connection",
    "JwtConnection",
    "JwtSuperuserConnection",
]

import logging
import sys
import time
from abc import abstractmethod
from typing import Any, Callable, Optional, Sequence, Set, Tuple, Union

import jwt
from requests import ConnectionError, Session
from requests_toolbelt import MultipartEncoder

from arango.exceptions import JWTAuthError, ServerConnectionError
from arango.http import HTTPClient
from arango.request import Request
from arango.resolver import HostResolver
from arango.response import Response
from arango.typings import Fields, Json

Connection = Union["BasicConnection", "JwtConnection", "JwtSuperuserConnection"]


class BaseConnection:
    """Base connection to a specific ArangoDB database."""

    def __init__(
        self,
        hosts: Fields,
        host_resolver: HostResolver,
        sessions: Sequence[Session],
        db_name: str,
        http_client: HTTPClient,
        serializer: Callable[..., str],
        deserializer: Callable[[str], Any],
    ) -> None:
        self._url_prefixes = [f"{host}/_db/{db_name}" for host in hosts]
        self._host_resolver = host_resolver
        self._sessions = sessions
        self._db_name = db_name
        self._http = http_client
        self._serializer = serializer
        self._deserializer = deserializer
        self._username: Optional[str] = None

    @property
    def db_name(self) -> str:
        """Return the database name.

        :returns: Database name.
        :rtype: str
        """
        return self._db_name

    @property
    def username(self) -> Optional[str]:
        """Return the username.

        :returns: Username.
        :rtype: str
        """
        return self._username

    def serialize(self, obj: Any) -> str:
        """Serialize the given object.

        :param obj: JSON object to serialize.
        :type obj: str | bool | int | float | list | dict | None
        :return: Serialized string.
        :rtype: str
        """
        return self._serializer(obj)

    def deserialize(self, string: str) -> Any:
        """De-serialize the string and return the object.

        :param string: String to de-serialize.
        :type string: str
        :return: De-serialized JSON object.
        :rtype: str | bool | int | float | list | dict | None
        """
        try:
            return self._deserializer(string)
        except (ValueError, TypeError):
            return string

    def prep_response(self, resp: Response, deserialize: bool = True) -> Response:
        """Populate the response with details and return it.

        :param deserialize: Deserialize the response body.
        :type deserialize: bool
        :param resp: HTTP response.
        :type resp: arango.response.Response
        :return: HTTP response.
        :rtype: arango.response.Response
        """
        if deserialize:
            resp.body = self.deserialize(resp.raw_body)
            if isinstance(resp.body, dict):
                resp.error_code = resp.body.get("errorNum")
                resp.error_message = resp.body.get("errorMessage")
                if resp.status_code == resp.error_code == 503:
                    raise ConnectionError  # Fallback to another host
        else:
            resp.body = resp.raw_body

        http_ok = 200 <= resp.status_code < 300
        resp.is_success = http_ok and resp.error_code is None
        return resp

    def process_request(
        self, host_index: int, request: Request, auth: Optional[Tuple[str, str]] = None
    ) -> Response:
        """Execute a request until a valid response has been returned.

        :param host_index: The index of the first host to try
        :type host_index: int
        :param request: HTTP request.
        :type request: arango.request.Request
        :return: HTTP response.
        :rtype: arango.response.Response
        """
        tries = 0
        indexes_to_filter: Set[int] = set()
        while tries < self._host_resolver.max_tries:
            try:
                resp = self._http.send_request(
                    session=self._sessions[host_index],
                    method=request.method,
                    url=self._url_prefixes[host_index] + request.endpoint,
                    params=request.params,
                    data=self.normalize_data(request.data),
                    headers=request.headers,
                    auth=auth,
                )

                return self.prep_response(resp, request.deserialize)
            except ConnectionError:
                url = self._url_prefixes[host_index] + request.endpoint
                logging.debug(f"ConnectionError: {url}")

                if len(indexes_to_filter) == self._host_resolver.host_count - 1:
                    indexes_to_filter.clear()
                indexes_to_filter.add(host_index)

                host_index = self._host_resolver.get_host_index(indexes_to_filter)
                tries += 1

        raise ConnectionAbortedError(
            f"Can't connect to host(s) within limit ({self._host_resolver.max_tries})"
        )

    def prep_bulk_err_response(self, parent_response: Response, body: Json) -> Response:
        """Build and return a bulk error response.

        :param parent_response: Parent response.
        :type parent_response: arango.response.Response
        :param body: Error response body.
        :type body: dict
        :return: Child bulk error response.
        :rtype: arango.response.Response
        """
        resp = Response(
            method=parent_response.method,
            url=parent_response.url,
            headers=parent_response.headers,
            status_code=parent_response.status_code,
            status_text=parent_response.status_text,
            raw_body=self.serialize(body),
        )
        resp.body = body
        resp.error_code = body["errorNum"]
        resp.error_message = body["errorMessage"]
        resp.is_success = False
        return resp

    def normalize_data(self, data: Any) -> Union[str, MultipartEncoder, None]:
        """Normalize request data.

        :param data: Request data.
        :type data: str | MultipartEncoder | None
        :return: Normalized data.
        :rtype: str | MultipartEncoder | None
        """
        if data is None:
            return None
        elif isinstance(data, (str, MultipartEncoder)):
            return data
        else:
            return self.serialize(data)

    def ping(self) -> int:
        """Ping the next host to check if connection is established.

        :return: Response status code.
        :rtype: int
        """
        request = Request(method="get", endpoint="/_api/collection")
        resp = self.send_request(request)
        if resp.status_code in {401, 403}:
            raise ServerConnectionError("bad username and/or password")
        if not resp.is_success:  # pragma: no cover
            raise ServerConnectionError(resp.error_message or "bad server response")
        return resp.status_code

    @abstractmethod
    def send_request(self, request: Request) -> Response:  # pragma: no cover
        """Send an HTTP request to ArangoDB server.

        :param request: HTTP request.
        :type request: arango.request.Request
        :return: HTTP response.
        :rtype: arango.response.Response
        """
        raise NotImplementedError


class BasicConnection(BaseConnection):
    """Connection to specific ArangoDB database using basic authentication.

    :param hosts: Host URL or list of URLs (coordinators in a cluster).
    :type hosts: [str]
    :param host_resolver: Host resolver (used for clusters).
    :type host_resolver: arango.resolver.HostResolver
    :param sessions: HTTP session objects per host.
    :type sessions: [requests.Session]
    :param db_name: Database name.
    :type db_name: str
    :param username: Username.
    :type username: str
    :param password: Password.
    :type password: str
    :param http_client: User-defined HTTP client.
    :type http_client: arango.http.HTTPClient
    """

    def __init__(
        self,
        hosts: Fields,
        host_resolver: HostResolver,
        sessions: Sequence[Session],
        db_name: str,
        username: str,
        password: str,
        http_client: HTTPClient,
        serializer: Callable[..., str],
        deserializer: Callable[[str], Any],
    ) -> None:
        super().__init__(
            hosts,
            host_resolver,
            sessions,
            db_name,
            http_client,
            serializer,
            deserializer,
        )
        self._username = username
        self._auth = (username, password)

    def send_request(self, request: Request) -> Response:
        """Send an HTTP request to ArangoDB server.

        :param request: HTTP request.
        :type request: arango.request.Request
        :return: HTTP response.
        :rtype: arango.response.Response
        """
        host_index = self._host_resolver.get_host_index()
        return self.process_request(host_index, request, auth=self._auth)


class JwtConnection(BaseConnection):
    """Connection to specific ArangoDB database using JWT authentication.

    :param hosts: Host URL or list of URLs (coordinators in a cluster).
    :type hosts: [str]
    :param host_resolver: Host resolver (used for clusters).
    :type host_resolver: arango.resolver.HostResolver
    :param sessions: HTTP session objects per host.
    :type sessions: [requests.Session]
    :param db_name: Database name.
    :type db_name: str
    :param username: Username.
    :type username: str
    :param password: Password.
    :type password: str
    :param http_client: User-defined HTTP client.
    :type http_client: arango.http.HTTPClient
    """

    def __init__(
        self,
        hosts: Fields,
        host_resolver: HostResolver,
        sessions: Sequence[Session],
        db_name: str,
        username: str,
        password: str,
        http_client: HTTPClient,
        serializer: Callable[..., str],
        deserializer: Callable[[str], Any],
    ) -> None:
        super().__init__(
            hosts,
            host_resolver,
            sessions,
            db_name,
            http_client,
            serializer,
            deserializer,
        )
        self._username = username
        self._password = password

        self.exp_leeway: int = 0
        self._auth_header: Optional[str] = None
        self._token: Optional[str] = None
        self._token_exp: int = sys.maxsize

        self.refresh_token()

    def send_request(self, request: Request) -> Response:
        """Send an HTTP request to ArangoDB server.

        :param request: HTTP request.
        :type request: arango.request.Request
        :return: HTTP response.
        :rtype: arango.response.Response
        """
        host_index = self._host_resolver.get_host_index()

        if self._auth_header is not None:
            request.headers["Authorization"] = self._auth_header

        resp = self.process_request(host_index, request)

        # Refresh the token and retry on HTTP 401 and error code 11.
        if resp.error_code != 11 or resp.status_code != 401:
            return resp

        now = int(time.time())
        if self._token_exp < now - self.exp_leeway:  # pragma: no cover
            return resp

        self.refresh_token()

        if self._auth_header is not None:
            request.headers["Authorization"] = self._auth_header

        return self.process_request(host_index, request)

    def refresh_token(self) -> None:
        """Get a new JWT token for the current user (cannot be a superuser).

        :return: JWT token.
        :rtype: str
        """
        request = Request(
            method="post",
            endpoint="/_open/auth",
            data={"username": self._username, "password": self._password},
        )

        host_index = self._host_resolver.get_host_index()

        resp = self.process_request(host_index, request)

        if not resp.is_success:
            raise JWTAuthError(resp, request)

        self._token = resp.body["jwt"]
        assert self._token is not None

        jwt_payload = jwt.decode(
            self._token,
            issuer="arangodb",
            algorithms=["HS256"],
            options={
                "require_exp": True,
                "require_iat": True,
                "verify_iat": True,
                "verify_exp": True,
                "verify_signature": False,
            },
        )
        self._token_exp = jwt_payload["exp"]
        self._auth_header = f"bearer {self._token}"


class JwtSuperuserConnection(BaseConnection):
    """Connection to specific ArangoDB database using superuser JWT.

    :param hosts: Host URL or list of URLs (coordinators in a cluster).
    :type hosts: [str]
    :param host_resolver: Host resolver (used for clusters).
    :type host_resolver: arango.resolver.HostResolver
    :param sessions: HTTP session objects per host.
    :type sessions: [requests.Session]
    :param db_name: Database name.
    :type db_name: str
    :param http_client: User-defined HTTP client.
    :type http_client: arango.http.HTTPClient
    :param superuser_token: User generated token for superuser access.
    :type superuser_token: str
    """

    def __init__(
        self,
        hosts: Fields,
        host_resolver: HostResolver,
        sessions: Sequence[Session],
        db_name: str,
        http_client: HTTPClient,
        serializer: Callable[..., str],
        deserializer: Callable[[str], Any],
        superuser_token: str,
    ) -> None:
        super().__init__(
            hosts,
            host_resolver,
            sessions,
            db_name,
            http_client,
            serializer,
            deserializer,
        )
        self._auth_header = f"bearer {superuser_token}"

    def send_request(self, request: Request) -> Response:
        """Send an HTTP request to ArangoDB server.

        :param request: HTTP request.
        :type request: arango.request.Request
        :return: HTTP response.
        :rtype: arango.response.Response
        """
        host_index = self._host_resolver.get_host_index()
        request.headers["Authorization"] = self._auth_header

        return self.process_request(host_index, request)
